#!/usr/bin/env python

# The MIT License (MIT)
#
# Copyright (c) 2015 Hordur K. Heidarsson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function

import os.path
import sys
import ast

import yaml

from .utils import *
from .metadata_writer import BagMetadataUtility
from .system_info_collector import SystemInfoCollector
from .config import *

import ConfigParser



def print_once(s):
    if not hasattr(print_once, 'd'):
        print_once.d = set()
    if not s in print_once.d:
        print(s)
        print_once.d.add(s)


def main():

    import argparse
    conf_parser = argparse.ArgumentParser(add_help=False)
    conf_parser.add_argument('-c', '--config', dest='config', type=str, help='Config file', default='~/.ros/rosbag_metadata.conf')
    args, remaining_argv = conf_parser.parse_known_args()

    bool_config_options = ('clean', 'write_rosbag_info', 'system_info', 'system_info_all', 'system_info_usb',
        'system_info_git', 'system_info_ros', 'system_info_env', 'system_info_full_env', 'system_info_ip', 'find_all',
        'debug', 'ask_template_defaults', 'extra_fields')
    allowed_config_options = bool_config_options + ('template',)

    config = {}
    default_fields = DEFAULT_FIELDS


    if args.config and os.path.exists(os.path.expanduser(args.config)):
        config_parser = ConfigParser.SafeConfigParser(allow_no_value=True)
        config_parser.read([os.path.expanduser(args.config)])
        config = dict(config_parser.items("config"))


        for k in config.keys():

            # Only allow certain options to be read from config
            if not k in allowed_config_options:
                del config[k]
                continue
            if k in bool_config_options:
                config[k] = config_parser.getboolean('config',k)

        if config_parser.has_section('default_fields'):
            default_fields = dict(config_parser.items("default_fields"))

            # Remove fields that conflict with system reserved fields
            for k in default_fields.keys():
                if k in SYSTEM_FIELDS:
                    del default_fields[k]

        # read config file

    parser = argparse.ArgumentParser(parents=[conf_parser,], description='Read or write metadata to bagfiles or textfiles.', formatter_class=argparse.RawDescriptionHelpFormatter,)

    parser.add_argument('path', metavar='path', type=str, help='Target (bagfile, yamlfile or directory)')

    mutexgroup = parser.add_mutually_exclusive_group()
    mutexgroup.add_argument('-r', '--read', dest='not_used', action='store_true', help='Only read metadata (default).')
    mutexgroup.add_argument('-w', '--write', dest='read', action='store_false', help='Write metadata.')

    # Write options
    writegroup = parser.add_argument_group('Write options')
    writegroup.add_argument('-t', '--template', dest='template', type=str, help='')
    writegroup.add_argument('--clean', dest='clean', action='store_true', help='Do not read existing metadata, start fresh.')
    writegroup.add_argument('--write-rosbag-info', dest='write_rosbag_info', action='store_true', help="Runs 'rosbag info --freq' on each bag files in the directory and saves along with metadata (does not apply to bagfile targets)")
    writegroup.add_argument('--ask-template-defaults', dest='ask_template_defaults', action='store_true', help='Ask for values of fields defined in template even if they have a default value.')
    writegroup.add_argument('--no-extra-fields', dest='extra_fields', action='store_false', help='Do not prompt for extra fields.')
    writegroup.add_argument('-y', '--yes', dest='no_prompt', action='store_true', help='Do not prompt for overwriting files.')

    systemgroup = parser.add_argument_group('System metadata options')
    systemgroup.add_argument('--no-system-info', dest='system_info', action='store_false', help='Do not collect any system info as part of written metadata')
    systemgroup.add_argument('--all-system-info', dest='system_info_all', action='store_true', help='Collect all system info for metadata (overrides specific options)')
    systemgroup.add_argument('--no-usb', dest='system_info_usb', action='store_false', help='Do not collect usb device info as part of system metadata')
    systemgroup.add_argument('--no-git', dest='system_info_git', action='store_false', help='Do not collect git repository info as part of written metadata')
    systemgroup.add_argument('--no-ros', dest='system_info_ros', action='store_false', help='Do not collect ros info as part of written metadata')
    systemgroup.add_argument('--no-env', dest='system_info_env', action='store_false', help='Do not collect environment variables as part of written metadata')
    systemgroup.add_argument('--no-full-env', dest='system_info_full_env', action='store_false', help='Only collect ROS environment variables as written metadata')
    systemgroup.add_argument('--no-ip', dest='system_info_ip', action='store_false', help='Do not collect IP information as part of written metadata')

    # Both options
    parser.add_argument('-a', '--find-all', dest='find_all', action='store_true', help='Searches for all possible metadata for given path (only applies to directory targets)')
    parser.add_argument('-d', '--debug', dest='debug', action='store_true', help='Output debug info')
    parser.add_argument('-v', '--version', action='version', version='rosbag_metadata %s' % VERSION)


    # Use data from config to set defaults
    parser.set_defaults(**config)

    args = parser.parse_args(remaining_argv)


    # bmu only uses non command-line options from the config, so we pass config directly (instead of vars(args))
    bmu = BagMetadataUtility(**config)

    found_data = bmu.extract(args.path, find_all=args.find_all)
    if len(found_data) == 0 or args.clean:
        existing_data = {}
    else:
        for d in found_data:
            pass # pick newest
        existing_data = found_data[0][1]

    skip_template_defaults = []

    if args.template:
        template = {}
        try:
            template = yaml.load(file(args.template, 'r'))
            for k in template.keys():
                if template[k] is not None:
                    skip_template_defaults.append(k)
            existing_data.update(template)
        except:
            print("Error loading template '%s'. Exiting." % args.template)
            raise
            exit(1)


    if args.read:
        for d in found_data:
            print('Found the following data in %s:' % d[0])
            print(yaml.dump(d[1]))
        exit(0)

    data = {}

    for k in existing_data.keys():
        if k in SYSTEM_FIELDS:
            del existing_data[k]
        if not args.ask_template_defaults and k in skip_template_defaults:
            if type(existing_data[k]) is dict and 'ask' in existing_data[k].keys() and existing_data[k]['ask']:
                existing_data[k] = existing_data[k]['value']
                continue
            print_once('\nUsing defaults from template:')
            print('%s=%s' % (k,existing_data[k]))
            data[k] = existing_data[k]
            del existing_data[k]

    try:
        if not type(default_fields) is dict:
            d = {}
            for k in default_fields:
                d[k] = None
            default_fields = d

        for k in set(tuple(default_fields.keys()) + tuple(existing_data.keys())):
            try:
                if not args.ask_template_defaults and k in skip_template_defaults:
                    continue

                print_once('\nUse Ctrl+D to leave field empty instead of using default value.')
                print_once("\nPlease fill out the following:")
                data[k] = query(k,default_fields.get(k) or existing_data.get(k))
            except EOFError:
                print('')
                pass

        if args.extra_fields:
            print('\n===Extra Fields===')

            while True:
                field = query('Field name')
                if field is not None and len(field) > 0:
                    field_data = query(field)
                    if field_data is not None and len(field_data) > 0:
                        data[field] = field_data
                    else:
                        break
                else:
                    break

    except (KeyboardInterrupt, EOFError):
        print('')
        pass
    except:
        raise

    if args.system_info:
        print('\nCollecting system/environment metadata')
        data[SYSTEM_INFO_FIELD] = SystemInfoCollector(**vars(args)).get_data()

    if args.write_rosbag_info and not bmu.is_bag_file(args.path):
        data[BAGS_INFO_FIELD] = bmu.get_rosbag_info(args.path)

    # Prune data of empty keys
    for k in data.keys():
        if data[k] is None or data[k] == '':
            del data[k]

    if args.debug:
        print('===============')
        print('About to write:')
        print(bmu.dict_to_yaml(data))
        print('===============')

    if args.no_prompt:
        overwrite_existing = True
    else:
        overwrite_existing = OVERWRITE_ASK
    bmu.write_metadata(args.path, data, overwrite_existing=overwrite_existing)

if __name__ == '__main__':
    main()



# The MIT License (MIT)
#
# Copyright (c) 2015 Hordur K. Heidarsson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import rosbag
import std_msgs
import rospy

import os.path
import yaml
import re
import datetime

from .config import *
from .utils import *

class BagMetadataUtility(object):
    """docstring for BagMetadataUtility"""
    def __init__(self, default_topic=DEFAULT_TOPIC, metadata_filename=METADATA_FILENAME, **kwargs):
        super(BagMetadataUtility, self).__init__()

        self.metadata_filename = metadata_filename
        self.default_topic = default_topic

    def is_bag_file(self, path):
        if path.endswith('.bag'): #assume it is a bagfile if it ends with .bag
            return True

        # Do some actual file inspection?
        # Open with rosbag? (does that take time if file is big?)

        return False

    def get_rosbag_info(self, path):
        res = {}
        path = os.path.normpath(os.path.join(os.getcwd(), path))
        if not os.path.isdir(path):
            path = os.path.dirname(path)
        for f in os.listdir(path):
            if f.endswith('.bag'):
                res[f] = self.get_info(f)
        return res


    def dict_to_yaml(self, data):
        data['_metadata_info'] = {'creator': 'rosbag_metadata', 'about': ABOUT, 'version': VERSION, 'url': URL, 'date': '%s' % datetime.datetime.now()}
        return yaml.dump(data)

    def write_metadata_file(self, filename, metadata_string, overwrite_existing=False):

        if os.path.exists(filename):
            if overwrite_existing == OVERWRITE_ASK: # Ask
                overwrite_existing = query_yes_no("Overwrite existing file '%s'?" % filename, 'n')
            if not overwrite_existing:
                return None
        with open(filename,'w') as f:
            f.write(metadata_string)
        return (filename, )

    def write_metadata(self, filename, metadata, overwrite_existing=False):

        if isinstance(metadata, dict): #convert dict to yaml
            metadata = self.dict_to_yaml(metadata)

        if os.path.isdir(filename):
            return self.write_metadata_file(os.path.join(filename, self.metadata_filename), metadata, overwrite_existing=overwrite_existing)
        elif not os.path.exists(filename):
            return self.write_metadata_file(filename, metadata, overwrite_existing=overwrite_existing)
        else:
            try:
                return self.inject_to_bag(filename, metadata)
            except rosbag.bag.ROSBagException, e:
                if e.value == 'This does not appear to be a bag file' or e.value == 'empty file':
                    if overwrite_existing:
                        return self.write_metadata_file(filename, metadata, overwrite_existing=overwrite_existing)
                else:
                    raise

        # Metadata not written
        return None


    def inject_to_bag(self, bagfile_name, metadata):
        with rosbag.Bag(bagfile_name, 'a') as bag:

            metadata_msg = std_msgs.msg.String(data=metadata)

            for _, _, t in bag.read_messages():
                break

            bag.write(self.default_topic, metadata_msg, t - rospy.rostime.Duration(0, 1))


    def find_split_files(self, bagfile_name):
        path = os.path.dirname(bagfile_name)
        info = get_info(bagfile_name, freq=False)

    def extract_from_bag(self, bagfile_name, use_yaml=True):
        with rosbag.Bag(bagfile_name, 'r') as bag:
            for msg_topic, msg, t in bag.read_messages(topics=[self.default_topic,]):
                if msg_topic == self.default_topic:
                    found = True
                    break
            if found:
                if use_yaml: # Try to parse data as yaml unless told not to
                    try:
                        return (bagfile_name, yaml.load(msg.data))
                    except:
                        pass
                return (bagfile_name, msg.data)
        return None

    def extract_from_dir(self, dirname, search_bags=True, find_all=False):
        res = []
        filename = os.path.join(dirname, self.metadata_filename)
        if os.path.exists(filename):
            res.append(self.extract_from_file(filename))
            if not find_all:
                return res

        if search_bags:
            for f in os.listdir(dirname):
                if f.endswith('.bag'):
                    r = self.extract_from_bag(f)
                    if r is not None:
                        res.append(r)
                        if not find_all:
                            return res

        return res

    def extract_from_file(self, filename):
        # try reading the file
        with open(filename, 'r') as f:
            data = f.read()
            return (filename, yaml.load(data))
        return None

    def extract(self, filename, use_yaml=True, find_all=False):
        found = False


        res = []

        if not os.path.exists(filename):
            return res

        # check if directory, then search for metadata files
        if os.path.isdir(filename):
            res = self.extract_from_dir(filename,find_all=find_all)
        else:
            try:
                res.append(self.extract_from_bag(filename, use_yaml=use_yaml))
            except rosbag.bag.ROSBagException, e:
                if e.value == 'This does not appear to be a bag file' or e.value == 'empty file':
                    res.append(self.extract_from_file(filename))
                else:
                    raise

        if res is None or len(res) == 0 or res[0] is None:
            return []

        return res

    def get_info(self, bagfile_name, freq=True):
        b = rosbag.Bag(bagfile_name, 'r',skip_index=not freq)
        return yaml.load(b._get_yaml_info())


    def get_full_info(self, bagfile_name, freq=True):
        return dict(self.get_info(bagfile_name, freq=freq).items() + self.extract(bagfile_name, use_yaml=True).items())



# The MIT License (MIT)
#
# Copyright (c) 2015 Hordur K. Heidarsson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# System data:
## git
### search through ros package path, search for git repositories, get version
## system
### uname -a
###

import os
import rospkg
from rospkg.common import PACKAGE_FILE
from rospkg.rospack import ManifestManager
import platform
import re
import subprocess
import git
import datetime
import socket
import netifaces

# http://stackoverflow.com/questions/8110310/simple-way-to-query-connected-usb-devices-info-in-python
def get_usb_devices():
    device_re = re.compile("Bus\s+(?P<bus>\d+)\s+Device\s+(?P<device>\d+).+ID\s(?P<id>\w+:\w+)\s(?P<tag>.+)$", re.I)
    df = subprocess.check_output("lsusb", shell=True)
    devices = []
    for i in df.split('\n'):
        if i:
            info = device_re.match(i)
            if info:
                dinfo = info.groupdict()
                dinfo['device'] = '/dev/bus/usb/%s/%s' % (dinfo.pop('bus'), dinfo.pop('device'))
                devices.append(dinfo)
    return devices

def get_ip_info():
    res = {}
    for iface in netifaces.interfaces():
        res[iface] = netifaces.ifaddresses(iface)
    return res

class SystemInfoCollector():
    def __init__(self, system_info_all=True, system_info_usb=True, system_info_git=True, system_info_ros=True, system_info_env=True, system_info_full_env=False, system_info_ip=True, **kwargs):

        self.use_usb = system_info_usb or system_info_all
        self.use_git = system_info_git or system_info_all
        self.use_ros = system_info_ros or system_info_all
        self.use_env = system_info_env or system_info_all
        self.use_full_env = system_info_full_env or system_info_all
        self.use_ip = system_info_ip or system_info_all

        self.rospack = rospkg.RosPack()
        self.rosstack = rospkg.RosStack()
        self.mm = ManifestManager(PACKAGE_FILE)

    def get_ros_package_version(self, stack_name):
        # copied from https://github.com/ros-infrastructure/rospkg/blob/master/scripts/rosversion
        try:
            version = self.rosstack.get_stack_version(stack_name)
        except rospkg.ResourceNotFound as e:
            try:
                # hack to make it work with wet packages
                path = self.mm.get_path(stack_name)
                package_manifest = os.path.join(path, 'package.xml')
                if os.path.exists(package_manifest):
                    from xml.etree.ElementTree import ElementTree
                    try:
                        root = ElementTree(None, package_manifest)
                        version = root.findtext('version')
                        return version
                    except Exception:
                        pass
            except rospkg.ResourceNotFound as e:
                return None
        return None

    def get_git_repo_info(self, path):
        res = {}

        repo = git.Repo(path)
        g = repo.git

        res['branch'] = repo.head.ref.name
        res['rev'] = g.rev_parse('HEAD')

        log = repo.head.ref.log()
        if len(log) > 0:
            last_log = log[-1]
            res['rev_time'] = datetime.datetime.fromtimestamp(last_log.time[0]).__str__()

        res['remotes'] = {}
        for r in repo.remotes:
            res['remotes'][r.name] = {'url': r.url}

        return res

    def find_git_repos(self, paths):
        res = {}
        for p in paths:
            if not os.path.exists(p):
                continue
            g = git.Git(p)
            try:
                git_root = g.rev_parse(show_toplevel=True)
                if not res.has_key(git_root):
                    res[git_root] = self.get_git_repo_info(git_root)
            except git.GitCommandError:
                pass
        return res


    def get_ros_info(self):
        res = {}

        rospack = rospkg.RosPack()

        if 'ROS_DISTRO' in os.environ:
            res['distro_name'] = os.environ['ROS_DISTRO']

        res['package_versions'] = {}
        for p in rospack.list():
            res['package_versions'][p] = self.get_ros_package_version(p)

        if self.use_git:
            res['git'] = self.find_git_repos(os.environ['ROS_PACKAGE_PATH'].split(':'))

        # search for git repositories within ros package path

        return res

    def get_data(self):
        res = {}
        if self.use_env:
            res['env'] = self.get_env(only_ros=not self.use_full_env)

        if self.use_ros:
            res['ros'] = self.get_ros_info()

        res['system'] = self.get_system_info()
        return res

    def get_env(self, only_ros=True):
        d = os.environ.copy()
        if only_ros:
            for k in d.keys():
                if not k.startswith('ROS_'):
                    del d[k]
                if k == 'ROS_PACKAGE_PATH':
                    d[k] = d[k].split(':')
        return d

    def get_system_info(self):
        res = {}
        res['platform'] = platform.platform()
        res['hostname'] = socket.gethostname()

        if self.use_ip:
            res['ip'] = get_ip_info()

        if self.use_usb:
            res['usb'] = get_usb_devices()
        return res
